@@ -2,7 +2,6 @@
 from __future__ import unicode_literals
 
 # python std lib
-import datetime
 import json
 import logging
 import random
@@ -45,11 +44,10 @@ from .utils import (
 from redis import Redis
 from redis.client import list_or_args, parse_info
 from redis.connection import Connection, SSLConnection
-from redis._compat import iteritems, basestring, izip, nativestr, long
+from redis._compat import iteritems, nativestr, long
 from redis.exceptions import (
     BusyLoadingError,
     ConnectionError,
-    DataError,
     RedisError,
     ResponseError,
     TimeoutError,
@@ -211,7 +209,7 @@ class RedisCluster(Redis):
         "ZCARD",
         "ZCOUNT",
         "ZRANGE",
-        "ZSCORE"
+        "ZSCORE",
     ]
 
     RESULT_CALLBACKS = dict_merge(
@@ -362,6 +360,11 @@ class RedisCluster(Redis):
 
             log.debug("Connection pool class " + str(connection_pool_cls))
 
+            # If connection pool fails to initialize, parent class (Redis) __del__
+            # will try to access self.connection before it's defined
+            # throwing an AttributeError.
+            self.connection = None
+
             pool = connection_pool_cls(
                 startup_nodes=startup_nodes,
                 init_slot_cache=init_slot_cache,
@@ -417,7 +420,7 @@ class RedisCluster(Redis):
             connection_pool_cls = ClusterConnectionPool
 
         connection_pool = connection_pool_cls.from_url(url, db=db, skip_full_coverage_check=skip_full_coverage_check, **kwargs)
-        
+
         if connection_pool.connection_class == SSLConnection:
             connection_pool.connection_class = SSLClusterConnection
 
@@ -442,7 +445,7 @@ class RedisCluster(Redis):
         """
         return ClusterPubSub(self.connection_pool, **kwargs)
 
-    def pipeline(self, transaction=None, shard_hint=None):
+    def pipeline(self, transaction=None, shard_hint=None, read_from_replicas=False):
         """
         Cluster impl:
             Pipelines do not work in cluster mode the same way they do in normal mode.
@@ -461,6 +464,7 @@ class RedisCluster(Redis):
             result_callbacks=self.result_callbacks,
             response_callbacks=self.response_callbacks,
             cluster_down_retry_attempts=self.cluster_down_retry_attempts,
+            read_from_replicas=read_from_replicas,
         )
 
     def transaction(self, *args, **kwargs):
@@ -487,7 +491,7 @@ class RedisCluster(Redis):
 
         if command in ['XREADGROUP', 'XREAD']:
             stream_idx = args.index(b'STREAMS')
-            keys_ids = list(args[stream_idx + 1: ])
+            keys_ids = list(args[stream_idx + 1:])
             idx_split = len(keys_ids) // 2
             keys = keys_ids[: idx_split]
             slots = {self.connection_pool.nodes.keyslot(key) for key in keys}
@@ -508,7 +512,7 @@ class RedisCluster(Redis):
         `res` is a dict with the following structure Dict(NodeName, CommandResult)
         """
         if command in self.result_callbacks:
-            return self.result_callbacks[command](command, res, **kwargs)
+            return self.result_callbacks[command](command, res)
 
         # Default way to handle result
         return first_key(command, res)
@@ -543,7 +547,7 @@ class RedisCluster(Redis):
 
         It will try the number of times specified by the config option "self.cluster_down_retry_attempts"
         which defaults to 3 unless manually configured.
-        
+
         If it reaches the number of times, the command will raises ClusterDownException.
         """
         for _ in range(0, self.cluster_down_retry_attempts):
@@ -587,6 +591,7 @@ class RedisCluster(Redis):
 
         while ttl > 0:
             ttl -= 1
+            connection = None
 
             try:
                 if asking:
@@ -636,10 +641,10 @@ class RedisCluster(Redis):
                 # This is the last attempt before we run out of TTL, raise the exception
                 if ttl == 1:
                     raise e
-            except (RedisClusterException, BusyLoadingError) as e:
+            except (RedisClusterException, BusyLoadingError):
                 log.exception("RedisClusterException || BusyLoadingError")
                 raise
-            except ConnectionError as e:
+            except ConnectionError:
                 log.exception("ConnectionError")
 
                 connection.disconnect()
@@ -661,9 +666,9 @@ class RedisCluster(Redis):
                     self.connection_pool.nodes.increment_reinitialize_counter(
                         count=self.connection_pool.nodes.reinitialize_steps,
                     )
-
-            except TimeoutError as e:
+            except TimeoutError:
                 log.exception("TimeoutError")
+                connection.disconnect()
 
                 if ttl < self.RedisClusterRequestTTL / 2:
                     time.sleep(0.05)
@@ -675,6 +680,7 @@ class RedisCluster(Redis):
                 self.connection_pool.disconnect()
                 self.connection_pool.reset()
                 self.refresh_table_asap = True
+                connection = None
 
                 raise e
             except MovedError as e:
@@ -689,7 +695,7 @@ class RedisCluster(Redis):
 
                 node = self.connection_pool.nodes.set_node(e.host, e.port, server_type='master')
                 self.connection_pool.nodes.slots[e.slot_id][0] = node
-            except TryAgainError as e:
+            except TryAgainError:
                 log.exception("TryAgainError")
 
                 if ttl < self.RedisClusterRequestTTL / 2:
@@ -698,8 +704,13 @@ class RedisCluster(Redis):
                 log.exception("AskError")
 
                 redirect_addr, asking = "{0}:{1}".format(e.host, e.port), True
+            except BaseException as e:
+                log.exception("BaseException")
+                connection.disconnect()
+                raise e
             finally:
-                self.connection_pool.release(connection)
+                if connection is not None:
+                    self.connection_pool.release(connection)
 
             log.debug("TTL loop : " + str(ttl))
 
@@ -726,7 +737,7 @@ class RedisCluster(Redis):
 
                 connection.send_command(*args)
                 res[node["name"]] = self.parse_response(connection, command, **kwargs)
-            except ClusterDownError as e:
+            except ClusterDownError:
                 self.connection_pool.disconnect()
                 self.connection_pool.reset()
                 self.refresh_table_asap = True
@@ -793,14 +804,19 @@ class RedisCluster(Redis):
             for slot in slots
         ]
 
-    def cluster_failover(self, node_id, option):
+    def cluster_failover(self, node_id, option=None):
         """
         Forces a slave to perform a manual failover of its master
 
         Sends to specefied node
         """
-        assert option.upper() in ('FORCE', 'TAKEOVER')  # TODO: change this option handling
-        return self.execute_command('CLUSTER FAILOVER', option, node_id=node_id)
+        if option:
+            if option.upper() not in ['FORCE', 'TAKEOVER']:
+                raise RedisError('Invalid option for CLUSTER FAILOVER command: {0}'.format(option))
+            else:
+                return self.execute_command('CLUSTER FAILOVER', option, node_id=node_id)
+        else:
+            return self.execute_command('CLUSTER FAILOVER', node_id=node_id)
 
     def cluster_info(self):
         """
